﻿/**
 * @file
 * @brief Renderer.hppの実装を記述する。
 * @author akino
 */
#include <XG3D/Renderer.hpp>

//------------------------------------------------------------
#include <XBase/ArrayLength.hpp>
#include <XBase/Console.hpp>
#include <XBase/Display.hpp>
#include <XBase/Matrix34.hpp>
#include <XBase/Matrix44.hpp>
#include <XBase/Pointer.hpp>
#include <XBase/Random.hpp>
#include "GLCMD.hpp"
#include "ShaderConstant.hpp"

#if defined(XLIBRARY_DEBUG)
    #include <stdlib.h>
#endif

//------------------------------------------------------------
namespace XG3D {
//------------------------------------------------------------
namespace {
    const GLchar tSHADER_SOURCE_VSH[] =
#if defined(XG3D_ENGINE_GL)
"#version 120\r\n"
#endif
"#ifdef GL_ES\r\n"
"    #define IN         attribute\r\n"
"    #define OUT        varying\r\n"
"    #define LOWP       lowp\r\n"
"    #define MEDIUMP    mediump\r\n"
"    precision MEDIUMP  float;\r\n"
"    precision MEDIUMP  int;\r\n"
"#else\r\n"
"    #define IN         attribute\r\n"
"    #define OUT        varying\r\n"
"    #define LOWP       \r\n"
"    #define MEDIUMP    \r\n"
"#endif\r\n"
""
"uniform mat4 uMtxProj;"
"uniform mat4 uMtxView;"
""
"IN  vec4 iVtxPosition;"
"IN  vec4 iVtxNormal;"
"IN  vec4 iVtxTexCoord;"
"IN  vec4 iVtxColor;"
"OUT vec4 vColor;"
"    "
"void main()"
"{"
"    gl_Position = uMtxProj * uMtxView * iVtxPosition;"
"    vColor = iVtxColor;"
"}"
"";
    const GLchar tSHADER_SOURCE_FSH[] =
#if defined(XG3D_ENGINE_GL)
"#version 120\r\n"
#endif
"#ifdef GL_ES\r\n"
"    #define IN         varying\r\n"
"    #define LOWP       lowp\r\n"
"    #define MEDIUMP    mediump\r\n"
"    precision MEDIUMP  float;\r\n"
"    precision MEDIUMP  int;\r\n"
"    precision LOWP     sampler2D;\r\n"
"#else\r\n"
"    #define IN         varying\r\n"
"    #define LOWP       \r\n"
"    #define MEDIUMP    \r\n"
"#endif\r\n"
""
"IN LOWP vec4 vColor;"
""
"void main()"
"{"
"    gl_FragColor = vColor;"
"}";
    bool tCreateShader( 
        GLuint* aShader 
        , const GLenum aShaderType
        , const GLchar* aSource 
        )
    {
        GLint status = GLint();
        XG3D_GLCMD( *aShader = glCreateShader( aShaderType ) );
        XG3D_GLCMD( glShaderSource( *aShader , 1 , &aSource , 0 ) );
        XG3D_GLCMD( glCompileShader( *aShader ) );                

#if defined(XLIBRARY_DEBUG)
        {
            GLint logLength;
            XG3D_GLCMD( glGetShaderiv(*aShader, GL_INFO_LOG_LENGTH, &logLength) );
            if (logLength > 0)
            {
                GLchar *log = (GLchar *)malloc(logLength);
                XG3D_GLCMD( glGetShaderInfoLog(*aShader, logLength, &logLength, log) );
                XBASE_COUTLINE( "%s" , log );
                free(log);
            }
        }
#endif
        
        XG3D_GLCMD( glGetShaderiv( *aShader , GL_COMPILE_STATUS , &status ) );
        if ( status == GL_FALSE )
        {
            XBASE_NOT_REACH_ASSERT();
            XG3D_GLCMD( glDeleteShader( *aShader ) );
            *aShader = GLuint();
            return false;
        }
        return true;
    }

    bool tLinkProgram( GLuint aProgram )
    {
        GLint status = GLint();
        XG3D_GLCMD( glLinkProgram( aProgram ) );
        XG3D_GLCMD( glGetProgramiv( aProgram , GL_LINK_STATUS , &status ) );
        if ( status == 0 )
        {
            XBASE_NOT_REACH_ASSERT();
            return false;
        }
        return true;
    }

    bool tValidateProgram( GLuint aProgram )
    {
        GLint status = GLint();
        XG3D_GLCMD( glValidateProgram( aProgram ) );
        XG3D_GLCMD( glGetProgramiv( aProgram , GL_VALIDATE_STATUS , &status ) );
        if ( status == 0 )
        {
            XBASE_NOT_REACH_ASSERT();
            return false;
        }
        return true;
    }

    bool tCreateShaderProgram( GLuint* aProgram )
    {
        // プログラム作成
        XG3D_GLCMD( *aProgram = glCreateProgram() );

        // シェーダーソースを作成
        GLuint srcVSH = GLuint();
        if ( !tCreateShader( &srcVSH , GL_VERTEX_SHADER , tSHADER_SOURCE_VSH ) )
        {
            XBASE_NOT_REACH_ASSERT();
            XG3D_GLCMD( glDeleteProgram( *aProgram ) );
            return false;
        }
        GLuint srcFSH = GLuint();
        if ( !tCreateShader( &srcFSH , GL_FRAGMENT_SHADER , tSHADER_SOURCE_FSH ) )
        {
            XBASE_NOT_REACH_ASSERT();
            XG3D_GLCMD( glDeleteShader( srcVSH ) );
            XG3D_GLCMD( glDeleteProgram( *aProgram ) );
            return false;
        }

        // アタッチ
        XG3D_GLCMD( glAttachShader( *aProgram , srcVSH ) );
        XG3D_GLCMD( glAttachShader( *aProgram , srcFSH ) );
        
        // 属性バインド
        XG3D_GLCMD( glBindAttribLocation( *aProgram , ShaderConstant::Attribute_Position , "iVtxPosition" ) );
        XG3D_GLCMD( glBindAttribLocation( *aProgram , ShaderConstant::Attribute_Normal   , "iVtxNormal" ) );
        XG3D_GLCMD( glBindAttribLocation( *aProgram , ShaderConstant::Attribute_TexCoord , "iVtxTexCoord" ) );
        XG3D_GLCMD( glBindAttribLocation( *aProgram , ShaderConstant::Attribute_Color    , "iVtxColor" ) );

        // リンク＆チェック
        if ( !tLinkProgram( *aProgram )
            || !tValidateProgram( *aProgram )
            )
        {
            XG3D_GLCMD( glDeleteShader( srcFSH ) );
            XG3D_GLCMD( glDeleteShader( srcVSH ) );
            XG3D_GLCMD( glDeleteProgram( *aProgram ) );
        }

        // 後始末
        XG3D_GLCMD( glDeleteShader( srcFSH ) );
        XG3D_GLCMD( glDeleteShader( srcVSH ) );

        // 成功
        return true;
    }
    // シングルトンインスタンス
    ::XBase::Pointer< Renderer > tInstance;
}
//------------------------------------------------------------
Renderer& Renderer::Instance()
{
    return tInstance.ref();
}

//------------------------------------------------------------
Renderer::Renderer( ::XBase::Display& aDisplay )
: mDisplay( aDisplay )
, mEXT()
{
    // セットアップ
    mEXT.setup( aDisplay );

    // シェーダープログラムの作成
    if ( !tCreateShaderProgram( &mEXT.shaderProgram ) )
    {
        XBASE_NOT_REACH_ASSERT();
        return;
    }

    // Uniform場所の取得
    mEXT.uniformLocations[ ShaderConstant::Uniform_MtxProj ] = glGetUniformLocation( mEXT.shaderProgram , "uMtxProj" ); 
    mEXT.uniformLocations[ ShaderConstant::Uniform_MtxView ] = glGetUniformLocation( mEXT.shaderProgram , "uMtxView" );

    // リセット
    reset();

    // インスタンス設定
    tInstance.set( *this );
}

//------------------------------------------------------------
Renderer::~Renderer()
{
    // インスタンス設定解除
    tInstance.unset( *this );

    // シェーダープログラムの破棄
    XG3D_GLCMD( glUseProgram( 0 ) );
    XG3D_GLCMD( glDeleteProgram( mEXT.shaderProgram ) );
    mEXT.shaderProgram = GLuint();
}

//------------------------------------------------------------
void Renderer::reset()
{    
    // viewport
	fbSetViewport( 0 , 0 , mDisplay.mainScreen().width() , mDisplay.mainScreen().height() );

    // fbClearColor
    fbSetClearColor( ::XBase::Color4::Zero() );

    // fbClearDepth
    fbSetClearDepth( 1.0F );

    // fbColorUpdate
    fbSetColorUpdate( true );
    
    // fbDepthUpdate
    fbSetDepthUpdate( true );

    // depthCompare
	// glDisable( GL_DEPTH_TEST );

    // mtxProjection
    sdSetMtxProjection( ::XBase::Matrix44::Ortho( -1 , 1 , 1 , -1 , 0 , 1 ) );

    // mtxView
    sdSetMtxModelView( ::XBase::Matrix34::Identity() );
}

//------------------------------------------------------------
void Renderer::fbSetViewport( 
    const uint aBaseX 
    , const uint aBaseY
    , const uint aWidth
    , const uint aHeight
    )
{
    XG3D_GLCMD( glViewport( GLint( aBaseX ) , GLint( aBaseY ) , GLint( aWidth ) , GLint( aHeight ) ) );
}

//------------------------------------------------------------
void Renderer::fbClear()
{
    GLbitfield bits = 0;
    if ( mEXT.colorUpdate )
    {
        bits |= GL_COLOR_BUFFER_BIT;
    }
    if ( mEXT.depthUpdate )
    {
        bits |= GL_DEPTH_BUFFER_BIT;
    }
    XG3D_GLCMD( glClear( bits ) );
}

//------------------------------------------------------------
void Renderer::fbSetClearColor( const ::XBase::Color4POD& aColor )
{
    XG3D_GLCMD( glClearColor( aColor.r , aColor.g , aColor.b , aColor.a ) );
}

//------------------------------------------------------------
void Renderer::fbSetClearDepth( const float aDepth )
{
#if defined(XG3D_ENGINE_GLES)
    XG3D_GLCMD( glClearDepthf( aDepth ) );
#else
    XG3D_GLCMD( glClearDepth( aDepth ) );
#endif
}

//------------------------------------------------------------
void Renderer::fbSetColorUpdate( const bool aIsEnable )
{
    const GLboolean val = aIsEnable ? GL_TRUE : GL_FALSE;
    XG3D_GLCMD( glColorMask( val , val , val , val ) );
    mEXT.colorUpdate = aIsEnable;
}

//------------------------------------------------------------
void Renderer::fbSetDepthUpdate( const bool aIsEnable )
{
    XG3D_GLCMD( glDepthMask( aIsEnable ? GL_TRUE : GL_FALSE ) );
    mEXT.depthUpdate = aIsEnable;
}

//------------------------------------------------------------
void Renderer::sdSetMtxProjection( const ::XBase::Matrix44POD& aMtx )
{
    XG3D_GLCMD( glUseProgram( mEXT.shaderProgram ) );
    XG3D_GLCMD( glUniformMatrix4fv(
        mEXT.uniformLocations[ ShaderConstant::Uniform_MtxProj ] 
        , 1
        , GL_FALSE
        , aMtx.v
        ));
}

//------------------------------------------------------------
void Renderer::sdSetMtxModelView( const ::XBase::Matrix34POD& aMtx )
{
    XG3D_GLCMD( glUseProgram( mEXT.shaderProgram ) );
    XG3D_GLCMD( glUniformMatrix4fv(
        mEXT.uniformLocations[ ShaderConstant::Uniform_MtxView ] 
        , 1
        , GL_FALSE
        , aMtx.toMatrix44().v
        ));
}

//------------------------------------------------------------
void Renderer::waitDrawDone()
{    
    XG3D_GLCMD( glFinish() );
}

//------------------------------------------------------------
void Renderer::copyToDisplay( ::XBase::Display& aDisplay )
{
    copyToScreen( aDisplay.mainScreen() );
}

//------------------------------------------------------------
Renderer_EXT::Renderer_EXT()
: shaderProgram( 0 )
, uniformLocations()
, colorUpdate( false )
, depthUpdate( false )
{	
    XBASE_STATIC_ASSERT( UNIFORM_COUNT == ShaderConstant::Uniform_TERMINATE );
}

//------------------------------------------------------------
}
//------------------------------------------------------------
// EOF
