﻿/**
 * @file
 * @brief RuntimeArray型を記述する。
 * @author akino
 */
#if defined(XBASE_INCLUDED_RUNTIMEARRAY_HPP)
#else
#define XBASE_INCLUDED_RUNTIMEARRAY_HPP

//------------------------------------------------------------
#include <XBase/IAllocator.hpp>
#include <XBase/NonCopyable.hpp>
#include <XBase/RuntimeAssert.hpp>

//------------------------------------------------------------
namespace XBase {
/// @addtogroup XBase-Collection
//@{
    /// 実行時に確保する大きさが決まる配列。
    template< typename T >
    class RuntimeArray : public ::XBase::NonCopyable
    {
    public:
        /// @name typedef
        //@{
        typedef T ValueType;
        //@}

        /// @name コンストラクタとデストラクタ
        //@{

        /**
         * @brief コンストラクタ。
         * @param aCount 配列長。
         * @param aAllocator 配列データを確保する際に使用するアロケータ。
         * @details 
         * 配列長が0の場合、アロケートは走りません。
         */
        RuntimeArray( uint aCount , IAllocator& aAllocator = IAllocator::Default() )
            : mAllocator( aAllocator )
            , mCount( aCount )
            , mPtr( 0 )
        {
            if ( 0 < mCount )
            {
                mPtr = static_cast< ValueType* >( mAllocator.alloc( sizeof( ValueType ) * mCount ) );
            }
        }

        /// デストラクタ。
        ~RuntimeArray()
        {
            if ( mPtr != 0 )
            {
                ValueType* ptr = mPtr;
                mPtr = 0;
                mAllocator.free( ptr );
            }
        }

        //@}

        /// @name アクセス
        //@{
        /// 配列長。
        uint count()const
        {
            return mCount;
        }

        /// 指定番目の要素にアクセス。
        ValueType& at( const uint aIndex )
        {
            if ( mCount <= aIndex )
            {
                XBASE_RANGE_ASSERT_MAX( aIndex , mCount );
                return mPtr[ 0 ]; // fail safe code
            }
            return mPtr[ aIndex ];
        }
        
        /// 指定番目の要素にアクセス。
        const ValueType& at( const uint aIndex )const
        {
            if ( mCount <= aIndex )
            {
                XBASE_RANGE_ASSERT_MAX( aIndex , mCount );
                return mPtr[ 0 ]; // fail safe code
            }
            return mPtr[ aIndex ];
        }

        //@}

        /// @name 演算子オーバーロード
        //@{
        ValueType& operator[]( const uint aIndex ) { return at( aIndex ); } ///< at() のエイリアス。
        const ValueType& operator[]( const uint aIndex )const { return at( aIndex ); } ///< at()const のエイリアス。
        //@}

    private:
        IAllocator& mAllocator;
        const uint  mCount;
        ValueType*  mPtr;
    };
//@}
}
//------------------------------------------------------------
#endif
// EOF
